"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const Token_1 = require("./Token");
const hooks_1 = require("./hooks");
const utils_1 = require("./utils");
exports.languages = new Map();
function tokenize(text, grammar) {
    if (!(grammar instanceof Map)) {
        throw new Error('A grammar must be a Map');
    }
    const startArray = [text];
    const rest = grammar.get('rest');
    if (rest instanceof Map) {
        for (const [token, def] of rest) {
            grammar.set(token, def);
        }
        //TODO: check if we need to shallow copy the grammar
        grammar.delete('rest');
    }
    tokenloop: for (const [token, def] of grammar) {
        const patterns = Array.isArray(def) ? def : [def];
        for (let j = 0; j < patterns.length; ++j) {
            const pattern = patterns[j];
            let lookbehindLength = 0;
            let regex = pattern instanceof RegExp ? pattern : pattern.pattern;
            const { inside, lookbehind, alias, greedy } = pattern;
            if (greedy && !regex.global) {
                const flags = (regex.toString().match(/[imuy]*$/) || [''])[0];
                regex = RegExp(regex.source, `${flags}g`);
            }
            for (let i = 0; i < startArray.length; i++) {
                // Don’t cache length as it changes during the loop
                const node = startArray[i];
                // Something went terribly wrong, ABORT, ABORT!
                if (startArray.length > text.length) {
                    break tokenloop;
                }
                if (node instanceof Token_1.Token) {
                    continue;
                }
                regex.lastIndex = 0;
                let match = regex.exec(node);
                if (!match) {
                    continue;
                }
                if (lookbehind) {
                    lookbehindLength = match[1].length;
                }
                const stringFrom = match.index - 1 + lookbehindLength;
                match = match[0].slice(lookbehindLength);
                const stringTo = stringFrom + match.length;
                const before = node.slice(0, stringFrom + 1);
                const after = node.slice(stringTo + 1);
                const inserts = [];
                if (before) {
                    inserts.push(before);
                }
                const wrapped = new Token_1.Token(token, inside ? tokenize(match, inside) : match, alias);
                inserts.push(wrapped);
                if (after) {
                    inserts.push(after);
                }
                startArray.splice(i, 1, ...inserts);
            }
        }
    }
    return startArray;
}
exports.tokenize = tokenize;
function highlight(text, language) {
    const grammar = exports.languages.get(language);
    if (!grammar) {
        throw new Error(`Your are trying to use ${language} language but its definition could not be found. Please make sure to use "addLanguage" to add your definition`);
    }
    const env = {
        grammar,
        language,
        code: text,
        highlightedCode: ''
    };
    hooks_1.run('before-highlight', env);
    const tokens = tokenize(env.code, env.grammar); // Specific for plugins
    const highlightedCode = Token_1.Token.stringify(utils_1.encode(tokens), language);
    env.highlightedCode = highlightedCode;
    hooks_1.run('after-highlight', env);
    return env.highlightedCode; // Specific for plugins
}
exports.highlight = highlight;
function addPlugin(plugin) {
    if (typeof plugin !== 'function') {
        throw new Error('Given Plugin must be a function');
    }
    plugin(hooks_1.add);
}
exports.addPlugin = addPlugin;
function addLanguage(name, def) {
    exports.languages.set(name, def);
}
exports.addLanguage = addLanguage;
function getLanguage(name) {
    return exports.languages.get(name);
}
exports.getLanguage = getLanguage;
